use bitvec::index::BitIdx;
use std::{error::Error, vec::Vec};

use deku::{
    bitvec::{BitStore, Msb0},
    ctx::ByteSize,
    deku_derive,
    reader::{Reader, ReaderRet},
    DekuError, DekuReader,
};

#[deku_derive(DekuRead)]
#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
#[deku(endian = "big")]
pub struct GroupId(u16);

trait MapToDekuParseError<T> {
    fn map_to_deku_parse_error(self) -> Result<T, DekuError>;
}

impl<T, E: Error> MapToDekuParseError<T> for Result<T, E> {
    fn map_to_deku_parse_error(self) -> Result<T, DekuError> {
        self.map_err(|e| DekuError::Parse(e.to_string().into()))
    }
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct DekuDataField {
    #[deku(reader = "read_size(deku::reader)")]
    data_size: ByteSize,

    #[deku(bytes_read = "data_size.0")]
    pub data: Vec<u8>,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct BlockHeader {
    pub group_id: GroupId,

    #[deku(reader = "read_size(deku::reader)")]
    pub data_size: ByteSize,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct DataBlock {
    pub header: BlockHeader,

    #[deku(bytes_read = "header.data_size.0")]
    pub data: Vec<DekuDataField>,
}

fn read_size<R: std::io::Read>(reader: &mut Reader<R>) -> Result<ByteSize, DekuError> {
    let first_byte: u8 = u8::from_reader_with_ctx(reader, ())?;

    let is_length_expanded = first_byte.get_bit::<Msb0>(BitIdx::new(0).map_to_deku_parse_error()?);

    match is_length_expanded {
        true => {
            let size_of_data_size: ByteSize = ByteSize((first_byte & 0b0111_1111) as usize);

            if size_of_data_size.0 > 4 {
                return Err(DekuError::Parse("Size of the length encoding is > 4, this is not normal. Probable parsing error".to_string().into()));
            };

            // maximum size of the buffer is 4, we use the offset to read values less than 4 bytes
            let buffer: &mut [u8; 4] = &mut [0; 4];
            let write_offset = 4 - size_of_data_size.0;

            match reader.read_bytes(size_of_data_size.0, &mut buffer[write_offset..])? {
                ReaderRet::Bits(_bit_vec) => Err(DekuError::Parse("Got bits when trying to read bytes -> reader is unaligned, this is not normal.".to_string().into())),
                ReaderRet::Bytes => Ok(ByteSize(u32::from_be_bytes(*buffer) as usize)),
            }
        }
        false => Ok(ByteSize(first_byte as usize)),
    }
}
